import subprocess
import atexit
import os
import sys
from subprocess import PIPE, TimeoutExpired, Popen
from typing import Optional, Tuple, List

""" Author: Ondrej Borysek, License: MIT, Last update: 2021-04-19"""


class ProcessHandler:
    """
       This simple wrapper for subprocess library enables easier Input/Output testing of other programs.
       No pip packages or additional linux packages are required.
       Should work on Linux, Windows, and hopefully also on MacOS.
       """

    def __init__(self, timeout: float) -> None:
        self.last_child : Optional[Popen[bytes]] = None
        self.__try_to_be_nice()
        self.timeout: float = timeout
        self.sigterm_grace_period = 0.2  # second
        atexit.register(self.__kill_children)
        # Using prctl would be safer, but less portable: https://pythonhosted.org/python-prctl/#prctl.set_pdeathsig

    def __try_to_be_nice(self) -> None:
        if getattr(os, 'nice', None):
            os.nice(10)

    def __kill_children(self) -> None:
        if self.last_child is None or self.last_child.poll() is not None:
            return

        print("Trying to terminate.")
        self.last_child.terminate()  # Ask the child to exit peacefully.
        
        try:
            self.last_child.wait(timeout=self.sigterm_grace_period)
        except TimeoutExpired:
            print("Child refused to terminate. Trying SIGKILL if available.")
            self.last_child.kill()  # Kill anyone still standing, just like Anakin did.
            # self.last_child.wait()  # Waiting would be safe on Linux, but on Windows .kill from Python is just .terminate

        if self.last_child.poll():
            try:
                _, _ = self.last_child.communicate(timeout=0.2)  # this can garbage collect the process
            except TimeoutExpired:
                print(f"[Warning] The process refused to die. You might need to kill the zombie manually.")
                ProcessHandler.linux_print_processes()

    def __start_process(self, user_command: List[str], input_str: Optional[str] = None) -> Tuple[int, str, str]:

        # Subprocess will be with the same niceness as the main program.
        # Warning: Do NOT use shell=True. That would only kill the shell, not the C program.
        proc = subprocess.Popen(user_command, shell=False, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        self.last_child = proc

        try:
            outb, errb = proc.communicate(input=input_str.encode("utf8") if input_str else None, timeout=self.timeout)
        except TimeoutExpired:
            print("Timeout, trying to kill.")
            proc.kill()
            outb, errb = proc.communicate()

        self.__kill_children()
        return proc.returncode, outb.decode(), errb.decode()

    @staticmethod
    def run(command: List[str],
            input_str: Optional[str] = None,
            print_output: bool = False,
            timeout: float = 5  # seconds
            ) -> Tuple[int, str, str]:
        if command is None:
            return -1, "", "No command provided"

        ph = ProcessHandler(timeout=timeout)
        return_code, outs, errs = ph.__start_process(command, input_str=input_str)
        if print_output:
            print("Command:", command)
            print("Input:\n---\n", input_str, sep="", end="---\n")
            print("Return code:", return_code)
            print("Output:\n---\n", ProcessHandler.prettyfi_the_output(outs), sep="", end="---\n")
            print("Error:\n---\n", ProcessHandler.prettyfi_the_output(errs), sep="", end="---\n\n")
        return return_code, outs, errs

    @staticmethod
    def prettyfi_the_output(a: str) -> str:
        return a.replace("\\r\\n", "\n").replace("\\n", "\n")

    @staticmethod
    def linux_print_processes() -> None:
        if not sys.platform.startswith('linux'):
            return
        print("""Please check that you don't see any zombie processes from the process of testing. They would have nice
              value of 19. You can use command "ps a -o pid,ni,time,cmd" """)


def usage_example() -> None:
    command_to_execute = ['/bin/sh', '-c', 'echo Lorem Ipsum']
    input_str = "Echo doesn't use input from stdin, but we can give it one anyway."
    return_code, outs, errs = ProcessHandler.run(command_to_execute, input_str=input_str, print_output=True)
    assert return_code == 0
    assert "Lorem Ipsum" in outs  # beware different newlines based on system
    assert len(errs) == 0


def usage_example_different_timeout() -> None:
    command_to_execute = ['/bin/sh', '-c', 'while true; do sleep 1; done;']
    return_code, outs, errs = ProcessHandler.run(command_to_execute, print_output=True, timeout=0.5)
    assert return_code == -9  # SIGKILL


if __name__ == "__main__":
    # Examples are presuming you're running Linux. If that's not the case, just edit the commands and arguments for
    # their Windows/other OS alternatives.
    usage_example()
    usage_example_different_timeout()



from astroid import nodes  # type: ignore
from typing import TYPE_CHECKING, Optional, List, Union, TypeVar

from pylint.checkers import BaseChecker  # type: ignore

if TYPE_CHECKING:
    from pylint.lint import PyLinter  # type: ignore

from edulint.linting.checkers.utils import \
    BaseVisitor, Named, get_name, get_assigned, is_any_assign, is_named, is_builtin


class AugmentAssignments(BaseChecker):  # type: ignore

    name = "augment-assignments"
    msgs = {
        "R6001": (
            "Use augmenting assignment: \"%s %s= %s\"",
            "use-augmenting-assignment",
            "Emitted when an assignment can be simplified by using its augmented version.",
        ),
    }

    def add_augmenting_message(self, bin_op: nodes.BinOp, param: nodes.BinOp, name: str) -> None:
        self.add_message("use-augmenting-assignment", node=bin_op, args=(name, bin_op.op, param.as_string()))

    def visit_binop(self, bin_op: nodes.BinOp) -> None:
        if not is_any_assign(bin_op.parent):
            return

        targets = get_assigned(bin_op.parent)
        if len(targets) != 1:
            return

        name = get_name(targets[0])
        if is_named(bin_op.left) and name == get_name(bin_op.left):
            self.add_augmenting_message(bin_op, bin_op.right, name)
        if bin_op.op in "+*|&" and is_named(bin_op.right) and name == get_name(bin_op.right):
            self.add_augmenting_message(bin_op, bin_op.left, name)


T = TypeVar("T")


class ImproveForLoop(BaseChecker):  # type: ignore

    name = "improve-for-loop"
    msgs = {
        "R6101": (
            "Iterate directly: \"for var in %s\" (with appropriate name for \"var\")",
            "use-foreach",
            "Emitted when a for-range loop is used while a for-each loop would suffice.",
        ),
        "R6102": (
            "Iterate using enumerate: \"for %s, var in enumerate(%s)\" (with appropriate name for \"var\")",
            "use-enumerate",
            "Emitted when a for-range loop is used with the element at each index is accessed as well.",
        ),
    }

    def __init__(self, linter: Optional["PyLinter"] = None) -> None:
        super().__init__(linter)

    def _is_for_range(self, node: nodes.For) -> bool:
        return isinstance(node.iter, nodes.Call) \
            and is_builtin(node.iter.func, "range") \
            and (len(node.iter.args) == 1
                 or (len(node.iter.args) == 2
                     and isinstance(node.iter.args[0], nodes.Const)
                     and node.iter.args[0].value == 0)) \
            and is_builtin(node.iter.args[0].func, "len") \
            and len(node.iter.args[0].args) == 1

    def _get_structure(self, node: nodes.For) -> nodes.NodeNG:
        return node.iter.args[0].args[0]

    class ModifiedListener(BaseVisitor[T]):

        def __init__(self, watched: List[Named]):
            self.watched = watched
            self.modified = {get_name(var): False for var in watched}
            super().__init__()

        def was_modified(self, node: nodes.NodeNG) -> bool:
            return self.modified[get_name(node)]

        @staticmethod
        def _is_assigned_to(node: Named) -> bool:
            return node in get_assigned(node.parent)

        @staticmethod
        def _is_same_var(var: Named, node: Named) -> bool:
            return var.scope() == node.scope() and isinstance(var, type(node)) and get_name(var) == get_name(node)

        def _visit_assigned_to(self, node: Named) -> T:
            if not self._is_assigned_to(node):
                return self.default

            for var in self.watched:
                if self._is_same_var(var, node):
                    self.modified[get_name(var)] = True

            return self.default

        def visit_name(self, name: nodes.Name) -> T:
            return self._visit_assigned_to(name)

        def visit_attribute(self, attribute: nodes.Attribute) -> T:
            return self._visit_assigned_to(attribute)

        def visit_assignname(self, assign: nodes.AssignName) -> T:
            return self._visit_assigned_to(assign)

    class StructureIndexedVisitor(ModifiedListener[bool]):
        default = False

        @staticmethod
        def combine(results: List[bool]) -> bool:
            return any(results)

        def __init__(self, structure: Union[nodes.Name, nodes.Attribute], index: nodes.Name):
            self.structure = structure
            self.index = index
            super().__init__([structure, index])

        def visit_subscript(self, subscript: nodes.Subscript) -> bool:
            sub_result = self.visit_many(subscript.get_children())
            if sub_result:
                return sub_result

            parent = subscript.parent
            if self.was_modified(self.structure) or self.was_modified(self.index):
                return False
            if not isinstance(subscript.value, type(self.structure)) \
                    or (isinstance(parent, nodes.Assign) and subscript in parent.targets) \
                    or not isinstance(subscript.slice, nodes.Name):
                return sub_result

            return subscript.slice.name == self.index.name and (
                (isinstance(self.structure, nodes.Name) and self.structure.name == subscript.value.name)
                or (isinstance(self.structure, nodes.Attribute) and self.structure.attrname == subscript.value.attrname)
            )

    class IndexUsedVisitor(ModifiedListener[bool]):
        default = False

        @staticmethod
        def combine(results: List[bool]) -> bool:
            return any(results)

        def __init__(self, structure: Union[nodes.Name, nodes.Attribute], index: nodes.Name):
            self.structure = structure
            self.index = index
            super().__init__([structure, index])

        def visit_name(self, name: nodes.Name) -> bool:
            super().visit_name(name)
            if name.name != self.index.name:
                return False
            if not isinstance(name.parent, nodes.Subscript) \
                    or not isinstance(self.structure, type(name.parent.value)) \
                    or self.modified[get_name(name)]:
                return True

            subscript = name.parent
            if not ((isinstance(subscript.value, nodes.Name)
                    and subscript.value.name == self.structure.name)
                    or (isinstance(subscript.value, nodes.Attirbute)
                    and subscript.value.attrname == self.structure.attrname)):
                return True

            return isinstance(subscript.parent, nodes.Assign) and subscript in subscript.parent.targets

    def visit_for(self, node: nodes.For) -> None:
        if not self._is_for_range(node):
            return

        structure = self._get_structure(node)
        index = node.target
        if not isinstance(structure, nodes.Name) and not isinstance(structure, nodes.Attribute):
            return

        if self.StructureIndexedVisitor(structure, node.target).visit_many(node.body):
            structure_name = get_name(structure)
            if self.IndexUsedVisitor(structure, node.target).visit_many(node.body):
                self.add_message("use-enumerate", args=(index.name, structure_name), node=node)
            else:
                self.add_message("use-foreach", args=structure_name, node=node)


def register(linter: "PyLinter") -> None:
    """This required method auto registers the checker during initialization.
    :param linter: The linter to register the checker to.
    """
    linter.register_checker(AugmentAssignments(linter))
    linter.register_checker(ImproveForLoop(linter))



from typing import List, Callable, Tuple, Dict, Set, Any
from edulint.config.arg import ImmutableArg
from edulint.linting.problem import ProblemJson, Problem
from edulint.linting.process_handler import ProcessHandler
from edulint.linting.overrides import get_overriders
from edulint.linting.tweakers import get_tweakers, Tweakers
from edulint.config.config import Config
from edulint.options import Option
from edulint.linters import Linter
import sys
import json
import pathlib


def flake8_to_problem(raw: ProblemJson) -> Problem:
    assert isinstance(raw["filename"], str), f'got {type(raw["filename"])} for filename'
    assert isinstance(raw["line_number"], int), f'got {type(raw["line_number"])} for line_number'
    assert isinstance(raw["column_number"], int), f'got {type(raw["column_number"])} for column_number'
    assert isinstance(raw["code"], str), f'got {type(raw["code"])} for code'
    assert isinstance(raw["text"], str), f'got {type(raw["text"])} for text'

    return Problem(
        Linter.FLAKE8,
        raw["filename"],
        raw["line_number"],
        raw["column_number"],
        raw["code"],
        raw["text"]
    )


def pylint_to_problem(raw: ProblemJson) -> Problem:
    assert isinstance(raw["path"], str), f'got {type(raw["path"])} for path'
    assert isinstance(raw["line"], int), f'got {type(raw["line"])} for line'
    assert isinstance(raw["column"], int), f'got {type(raw["column"])} for column'
    assert isinstance(raw["message-id"], str), f'got {type(raw["message-id"])} for message-id'
    assert isinstance(raw["message"], str), f'got {type(raw["message"])} for message'
    assert isinstance(raw["endLine"], int) or raw["endLine"] is None, f'got {type(raw["endLine"])} for endLine'
    assert isinstance(raw["endColumn"], int) or raw["endColumn"] is None, f'got {type(raw["endColumn"])} for endColumn'

    return Problem(
        Linter.PYLINT,
        raw["path"],
        raw["line"],
        raw["column"],
        raw["message-id"],
        raw["message"],
        raw["endLine"],
        raw["endColumn"]
    )


def lint_any(
        linter: Linter, filenames: List[str], linter_args: List[str], config_arg: ImmutableArg,
        result_getter: Callable[[Any], Any],
        out_to_problem: Callable[[ProblemJson], Problem]) -> List[Problem]:
    command = [sys.executable, "-m", str(linter)] + linter_args + list(config_arg.val) + filenames  # type: ignore
    return_code, outs, errs = ProcessHandler.run(command, timeout=10)

    print(errs, file=sys.stderr, end="")
    if (linter == Linter.FLAKE8 and return_code not in (0, 1)) or (linter == Linter.PYLINT and return_code == 32):
        print(f"edulint: {linter} exited with {return_code}", file=sys.stderr)
        exit(return_code)

    if not outs:
        return []

    result = result_getter(json.loads(outs))
    return list(map(out_to_problem, result))


def lint_flake8(filenames: List[str], config: Config) -> List[Problem]:
    flake8_args = ["--format=json"]
    return lint_any(
        Linter.FLAKE8, filenames, flake8_args, config[Option.FLAKE8],
        lambda r: [problem for problems in r.values() for problem in problems],
        flake8_to_problem)


def lint_pylint(filenames: List[str], config: Config) -> List[Problem]:
    cwd = pathlib.Path(__file__).parent.resolve()
    pylint_args = [f'--rcfile={cwd}/.pylintrc', "--output-format=json"]
    return lint_any(
        Linter.PYLINT, filenames, pylint_args, config[Option.PYLINT],
        lambda r: r, pylint_to_problem)


def apply_overrides(problems: List[Problem], overriders: Dict[str, Set[str]]) -> List[Problem]:
    codes_on_lines: Dict[int, Set[str]] = {line: set() for line in set([problem.line for problem in problems])}

    for problem in problems:
        codes_on_lines[problem.line].add(problem.code)

    result = []
    for problem in problems:
        o = overriders.get(problem.code, set())
        if not (o & codes_on_lines[problem.line]):
            result.append(problem)

    return result


def apply_tweaks(problems: List[Problem], tweakers: Tweakers, config: Config) -> List[Problem]:
    result = []
    for problem in problems:
        tweaker = tweakers.get((problem.source, problem.code))
        if tweaker:
            if tweaker.should_keep(problem, [arg for arg in config if arg.option in tweaker.used_options]):
                problem.text = tweaker.get_reword(problem)
                result.append(problem)
        else:
            result.append(problem)
    return result


def lint_one(filename: str, config: Config) -> List[Problem]:
    return lint([filename], config)


def sort(filenames: List[str], problems: List[Problem]) -> List[Problem]:
    indices = {fn: i for i, fn in enumerate(filenames)}
    problems.sort(key=lambda problem: (indices[problem.path], problem.line, problem.column))
    return problems


def lint(filenames: List[str], config: Config) -> List[Problem]:
    result = lint_flake8(filenames, config) + lint_pylint(filenames, config)
    result = apply_overrides(result, get_overriders())
    result = apply_tweaks(result, get_tweakers(), config)
    return sort(filenames, result)


def lint_many(partition: List[Tuple[List[str], Config]]) -> List[Problem]:
    return [problem for filenames, config in partition for problem in lint(filenames, config)]



from edulint.config.arg import ProcessedArg, UnprocessedArg, ImmutableArg
from edulint.options import UnionT, ImmutableT, Option, TakesVal, OptionParse, get_option_parses, get_name_to_option
from edulint.config.config_translations import get_config_translations, get_ib111_translations, Translation
from typing import Dict, List, Optional, Tuple, Iterator
from dataclasses import dataclass
from argparse import Namespace
import re
import sys
import shlex


@dataclass(frozen=True)
class Config:

    config: Tuple[ImmutableArg]

    @staticmethod
    def to_immutable(v: UnionT) -> ImmutableT:
        return v if not isinstance(v, list) else tuple(v)

    def __init__(self, config: Optional[List[ProcessedArg]] = None,
                 option_parses: Dict[Option, OptionParse] = get_option_parses()) -> None:
        config = config if config is not None else []
        wip_config: List[Optional[ImmutableArg]] = [None for _ in Option]

        for arg in config:
            assert wip_config[int(arg.option)] is None
            wip_config[int(arg.option)] = ImmutableArg(arg.option, self.to_immutable(arg.val))

        object.__setattr__(self, "config", tuple([arg if arg is not None else ImmutableArg(
            o, self.to_immutable(option_parses[o].default)) for o, arg in zip(Option, wip_config)]))

    def __str__(self) -> str:
        return f"Config({', '.join(arg.option.name + '=' + str(arg.val) for arg in self.config)})"

    def __getitem__(self, option: Option) -> ImmutableArg:
        return self.config[int(option)]

    def __contains__(self, option: Option) -> bool:
        return self[option] is not None

    def __iter__(self) -> Iterator[ImmutableArg]:
        return filter(lambda x: x is not None, self.config.__iter__())


def extract_args(filename: str) -> List[str]:
    edulint_re = re.compile(r"\s*#[\s#]*edulint:\s*", re.IGNORECASE)
    ib111_re = re.compile(r"\s*from\s+ib111\s+import\s+week_(\d+)", re.IGNORECASE)

    result: List[str] = []
    with open(filename) as f:
        for i, line in enumerate(f):
            line = line.strip()

            edmatch = edulint_re.match(line)
            if edmatch:
                raw_args = line[edmatch.end():]
                result.extend(shlex.split(raw_args))

            ibmatch = ib111_re.match(line)
            if ibmatch:
                result.append(f"{Option.IB111_WEEK.to_name()}={ibmatch.group(1)}")

    return result


def parse_args(args: List[str], option_parses: Dict[Option, OptionParse]) -> List[UnprocessedArg]:
    name_to_option = get_name_to_option(option_parses)

    def get_name_val(arg: str) -> Tuple[str, Optional[str]]:
        if "=" in arg:
            name, val = arg.split("=", 1)
            return name, val
        return arg, None

    result: List[UnprocessedArg] = []
    for arg in args:
        name, val = get_name_val(arg)
        option = name_to_option.get(name)

        if option is None:
            print(f"edulint: unrecognized option {name}", file=sys.stderr)
        else:
            option_parse = option_parses[option]
            if option_parse.takes_val == TakesVal.YES and val is None:
                print(f"edulint: option {name} takes an argument but none was supplied", file=sys.stderr)
            elif option_parse.takes_val == TakesVal.NO and val is not None:
                print(f"edulint: option {name} takes no argument but {val} was supplied", file=sys.stderr)
            else:
                result.append(UnprocessedArg(option, val))

    return result


def fill_in_val(arg: UnprocessedArg, translation: List[str]) -> List[str]:
    result = []
    for t in translation:
        if "<val>" in t:
            assert isinstance(arg.val, str)
            result.append(t.replace("<val>", arg.val))
        else:
            result.append(t)
    return result


def combine_and_translate(
        args: List[UnprocessedArg],
        option_parses: Dict[Option, OptionParse],
        config_translations: Dict[Option, Translation],
        ib111_translations: List[Translation]) -> Config:

    def combine(option_vals: List[UnionT], option: Option, val: Optional[str]) -> None:
        parse = option_parses[option]
        old_val = option_vals[int(option)]
        option_vals[int(option)] = parse.combine(old_val, parse.convert(val))

    def apply_translation(option_vals: List[UnionT], translated: Translation) -> None:
        translated_option = translated.for_linter.to_option()
        for val in translated.vals:
            combine(option_vals, translated_option, val)

    option_vals = [option_parses[o].default for o in Option]

    for arg in args:
        combine(option_vals, arg.option, arg.val)

        translated = config_translations.get(arg.option)
        if translated is not None:
            apply_translation(option_vals, translated)

    ib111_week = option_vals[int(Option.IB111_WEEK)]
    if ib111_week is not None:
        assert isinstance(ib111_week, int)
        if 0 <= ib111_week < len(ib111_translations):
            apply_translation(option_vals, ib111_translations[ib111_week])
        else:
            print(f"edulint: option {Option.IB111_WEEK.to_name()} has value {ib111_week} which is invalid;"
                  f"allowed values are 0 to {len(ib111_translations)}", file=sys.stderr)

    return Config([ProcessedArg(o, v) for o, v in zip(Option, option_vals) if v is not None])


def get_config(
        filename: str, cmd_args: List[str],
        option_parses: Dict[Option, OptionParse] = get_option_parses(),
        config_translations: Dict[Option, Translation] = get_config_translations(),
        ib111_translation: List[Translation] = get_ib111_translations()) -> Config:
    extracted = extract_args(filename) + cmd_args
    parsed = parse_args(extracted, option_parses)
    return combine_and_translate(parsed, option_parses, config_translations, ib111_translation)


def get_cmd_args(args: Namespace) -> List[str]:
    return [s for arg in args.options for s in shlex.split(arg)]



from edulint.linters import Linter
from edulint.options import Option
from typing import Dict, List
from dataclasses import dataclass


@dataclass(frozen=True)
class Translation:
    for_linter: Linter
    vals: List[str]


CONFIG_TRANSLATIONS: Dict[Option, Translation] = {
    Option.ENHANCEMENT: Translation(
        Linter.PYLINT,
        ["--enable=no-self-use,superfluous-parens,consider-using-min-builtin,"
         "consider-using-max-builtin,consider-using-with,unspecified-encoding,"
         "use-augmenting-assignment"
         ]
    ),
    Option.PYTHON_SPEC: Translation(
        Linter.PYLINT,
        ["--enable=unidiomatic-typecheck,misplaced-format-function,"
         "unnecessary-lambda,protected-access,multiple-imports,"
         "wrong-import-position,consider-using-from-import,wildcard-import,"
         "reimported,improve-for-loop,consider-iterating-dictionary,"
         "consider-using-dict-items,consider-using-f-string,"
         "inconsistent-return-statements,consider-swap-variables,"
         "consider-using-join,consider-using-set-comprehension,"
         "unnecessary-comprehension,use-a-generator,use-list-literal,"
         "use-dict-literal,invalid-name,consider-using-in"]
    ),
    Option.ALLOWED_ONECHAR_NAMES: Translation(
        Linter.PYLINT,
        ["--bad-names-rgxs=^[a-z]$"]
    )
}


def get_config_translations() -> Dict[Option, Translation]:
    return CONFIG_TRANSLATIONS


IB111_WEEK_TRANSLATIONS: List[Translation] = [
    Translation(Linter.PYLINT, ["--disable=consider-using-in,consider-swap-variables"]),  # 0
    Translation(Linter.PYLINT, ["--disable=consider-using-in,consider-swap-variables"]),  # 1
    Translation(Linter.PYLINT, ["--disable=consider-using-in,consider-swap-variables"]),  # 2
    Translation(Linter.PYLINT, ["--disable=consider-using-in"]),  # 3
    Translation(Linter.PYLINT, ["--disable=consider-using-in"]),  # 4
    Translation(Linter.PYLINT, ["--disable=consider-using-in"]),  # 5
    Translation(Linter.PYLINT, ["--disable=consider-using-in"]),  # 6
    Translation(Linter.PYLINT, []),  # 7
    Translation(Linter.PYLINT, []),  # 8
    Translation(Linter.PYLINT, []),  # 9
    Translation(Linter.PYLINT, []),  # 10
    Translation(Linter.PYLINT, []),  # 11
    Translation(Linter.PYLINT, []),  # 12
]


def get_ib111_translations() -> List[Translation]:
    assert len(IB111_WEEK_TRANSLATIONS) == 13
    return IB111_WEEK_TRANSLATIONS



from enum import Enum, auto
from edulint.options import Option


class Linter(Enum):
    EDULINT = auto()
    PYLINT = auto()
    FLAKE8 = auto()

    def __str__(self: Enum) -> str:
        return self.name.lower()

    @staticmethod
    def from_name(linter_str: str) -> "Linter":
        for linter in Linter:
            if str(linter) == linter_str.lower():
                return linter
        assert False, "no such linter: " + linter_str

    def to_name(self) -> str:
        return str(self)

    @classmethod
    def from_option(cls, option: Option) -> "Linter":
        return cls.from_name(option.to_name())

    def to_option(self) -> Option:
        return Option.from_name(self.to_name())



from typing import Dict, Set


OVERRIDERS = {
    "C0104": {"R6101", "R6001", "F841", "W0613"},
    # C0104=diallowed-name; R6101=use-for-each; R6001=augmenting-assignment
    #                       F841=unused variable; W0613=unused-argument
    "C0103": {"R6101", "R6001", "F841", "W0613"},
    # C0103=invalid-name
    "E225": {"R6001"},
    # E225=missing whitespace around operator
    "R1705": {"R1703"},
    # R1705=no-else-return, R1703=simplifiable-if-statement
}


def get_overriders() -> Dict[str, Set[str]]:
    return OVERRIDERS



from edulint.linters import Linter
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config
from marshmallow import fields
from typing import Optional, Dict, Union

ProblemJson = Dict[str, Union[str, int]]


@dataclass_json
@dataclass
class Problem:
    source: Linter = field(metadata=config(
        encoder=Linter.to_name,
        decoder=Linter.from_name,
        mm_field=fields.Str()
    ))
    path: str
    line: int
    column: int
    code: str
    text: str
    end_line: Optional[int] = None
    end_column: Optional[int] = None

    def set_source(self, v: Linter) -> "Problem":
        self.source = v
        return self

    def set_path(self, v: str) -> "Problem":
        self.path = v
        return self

    def set_line(self, v: int) -> "Problem":
        self.line = v
        return self

    def set_column(self, v: int) -> "Problem":
        self.column = v
        return self

    def set_code(self, v: str) -> "Problem":
        self.code = v
        return self

    def set_text(self, v: str) -> "Problem":
        self.text = v
        return self

    def set_end_line(self, v: Optional[int]) -> "Problem":
        self.end_line = v
        return self

    def set_end_column(self, v: Optional[int]) -> "Problem":
        self.end_column = v
        return self

    def __str__(self) -> str:
        return f"{self.path}:{self.line}:{self.column}: " \
               f"{self.code} {self.text}"



from typing import Any, TypeVar, Generic, List, Iterable, Union, Optional, cast
from astroid import nodes  # type: ignore
import sys
import inspect
from pylint.checkers import utils  # type: ignore


def eprint(*args: Any, **kwargs: Any) -> None:
    print(*args, file=sys.stderr, **kwargs)


T = TypeVar("T")


class BaseVisitor(Generic[T]):
    default: T = None  # type: ignore

    @classmethod
    def combine(cls, results: List[T]) -> T:
        if not results:
            return cls.default  # type: ignore
        return results[-1]

    def visit(self, node: nodes.NodeNG) -> T:
        return node.accept(self)  # type: ignore

    def visit_many(self, nodes: Iterable[nodes.NodeNG]) -> T:
        return self.combine([self.visit(node) for node in nodes])


def generic_visit(self: BaseVisitor[T], node: nodes.NodeNG) -> T:
    return self.combine([child.accept(self) for child in node.get_children()])


for name, obj in inspect.getmembers(nodes, inspect.isclass):
    setattr(BaseVisitor, f"visit_{obj.__name__.lower()}", generic_visit)


# rightfully stolen from
# https://github.com/PyCQA/pylint/blob/ca80f03a43bc39e4cc2c67dc99817b3c9f13b8a6/pylint/checkers/refactoring/recommendation_checker.py
def is_builtin(node: nodes.NodeNG, function: Optional[str] = None) -> bool:
    inferred = utils.safe_infer(node)
    if not inferred:
        return False
    return utils.is_builtin_object(inferred) and function is None or inferred.name == function


def is_multi_assign(node: nodes.NodeNG) -> bool:
    return hasattr(node, "targets")


def is_assign(node: nodes.NodeNG) -> bool:
    return hasattr(node, "target")


def is_any_assign(node: nodes.NodeNG) -> bool:
    return is_assign(node) or is_multi_assign(node)


def get_assigned(node: nodes.NodeNG) -> List[nodes.NodeNG]:
    if is_multi_assign(node):
        return cast(List[nodes.NodeNG], node.targets)
    if is_assign(node):
        return [node.target]
    return []


Named = Union[nodes.Name, nodes.Attribute, nodes.AssignName]


def is_named(node: nodes.NodeNG) -> bool:
    return hasattr(node, "name") or hasattr(node, "attrname")


def get_name(node: Named) -> str:
    return str(node.name) if hasattr(node, "name") else f".{node.attrname}"



from .linters import Linter
from .config.config import Config
from .linting.problem import Problem
from .linting.linting import lint_one, lint_many

__all__ = ["Linter", "Config", "Problem", "lint_one", "lint_many"]

__version__ = "1.0.0"
__version_info__ = tuple(map(int, __version__.split(".")))



from dataclasses import dataclass
from typing import Dict, List, Any, TypeVar, Optional, Union, Callable, Tuple
from enum import Enum, auto


class NumberFromZero(Enum):
    def __new__(cls, *args: Any) -> "NumberFromZero":
        value = len(cls.__members__)
        obj = object.__new__(cls)
        obj._value_ = value
        return obj


T = TypeVar("T")
UnionT = Union[bool, List[str], Optional[str], Optional[int]]
ImmutableT = Union[bool, Tuple[str, ...], Optional[str], Optional[int]]


class Option(NumberFromZero):

    PYLINT = ()
    FLAKE8 = ()
    ENHANCEMENT = ()
    PYTHON_SPEC = ()
    ALLOWED_ONECHAR_NAMES = ()
    IB111_WEEK = ()

    def to_name(self) -> str:
        return self.name.lower().replace("_", "-")

    @staticmethod
    def from_name(option_str: str) -> "Option":
        for option in Option:
            if option.to_name() == option_str.lower():
                return option
        assert False, "no such option: " + option_str

    def __int__(self) -> int:
        return self.value  # type: ignore


class TakesVal(Enum):
    YES = auto()
    NO = auto()
    OPTIONAL = auto()


class MultivaluedEnum(Enum):
    def __new__(cls, *args: Any, **kwds: Any) -> Any:
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj


class Type(MultivaluedEnum):
    @staticmethod
    def _to_bool_val(val: Optional[str]) -> bool:
        return val is None or val in ("true", "on")

    @staticmethod
    def _to_list_val(val: Optional[str]) -> List[str]:
        return [val] if val is not None else []

    @staticmethod
    def _to_str_val(val: Optional[str]) -> Optional[str]:
        return val

    @staticmethod
    def _to_int_val(val: Optional[str]) -> Optional[int]:
        return int(val) if val is not None and val.isdecimal() else None

    def __init__(self, _: Enum, convert: Callable[[Optional[str]], UnionT]):
        self.convert: Callable[[Optional[str]], UnionT] = convert.__func__  # type: ignore

    def __call__(self, arg: Optional[str]) -> UnionT:
        return self.convert(arg)

    BOOL = (auto(), _to_bool_val)
    LIST = (auto(), _to_list_val)
    STR = (auto(), _to_str_val)
    INT = (auto(), _to_int_val)


class Combine(MultivaluedEnum):
    @staticmethod
    def _keep_right_combine(_lt: T, rt: T) -> T:
        return rt

    @staticmethod
    def _append_combine(lt: List[T], rt: T) -> List[T]:
        return lt + [rt]

    @staticmethod
    def _extend_combine(lt: List[T], rt: List[T]) -> List[T]:
        return lt + rt

    def __init__(self, _: Enum, combine: Callable[[UnionT, UnionT], UnionT]):
        self.combine: Callable[[UnionT, UnionT], UnionT] = combine.__func__  # type: ignore

    def __call__(self, lt: UnionT, rt: UnionT) -> UnionT:
        return self.combine(lt, rt)

    REPLACE = (auto(), _keep_right_combine)
    APPEND = (auto(), _append_combine)
    EXTEND = (auto(), _extend_combine)


@dataclass
class OptionParse:
    option: Option
    help_: str
    takes_val: TakesVal
    default: UnionT
    convert: Type
    combine: Combine


OPTIONS: List[OptionParse] = [
    OptionParse(
        Option.PYLINT,
        "arguments to be passed to pylint",
        TakesVal.YES,
        [],
        Type.LIST,
        Combine.EXTEND
    ),
    OptionParse(
        Option.FLAKE8,
        "arguments to be passed to flake8",
        TakesVal.YES,
        [],
        Type.LIST,
        Combine.EXTEND
    ),
    OptionParse(
        Option.ENHANCEMENT,
        "enable checking for ways to improve the code further",
        TakesVal.NO,
        False,
        Type.BOOL,
        Combine.REPLACE
    ),
    OptionParse(
        Option.PYTHON_SPEC,
        "enable checking for ways to improve the code with Python-specific constructions",
        TakesVal.NO,
        False,
        Type.BOOL,
        Combine.REPLACE
    ),
    OptionParse(
        Option.ALLOWED_ONECHAR_NAMES,
        "only listed characters are allowed to be variable names of length one",
        TakesVal.YES,
        None,
        Type.STR,
        Combine.REPLACE
    ),
    OptionParse(
        Option.IB111_WEEK,
        "set which week's limitation should be applied",
        TakesVal.YES,
        None,
        Type.INT,
        Combine.REPLACE
    )
]


def get_option_parses() -> Dict[Option, OptionParse]:
    return {parse.option: parse for parse in OPTIONS}


def get_name_to_option(option_parses: Dict[Option, OptionParse]) -> Dict[str, Option]:
    return {parse.option.to_name(): opt for opt, parse in option_parses.items()}



from edulint.linters import Linter
from edulint.options import Option
from edulint.config.arg import ImmutableArg
from edulint.linting.problem import Problem
from dataclasses import dataclass
from typing import Callable, Optional, Pattern, Match, AnyStr, Dict, Tuple, Set, List
import re


@dataclass
class Tweaker:
    used_options: Set[Option]

    pattern: Pattern[AnyStr]  # type: ignore
    keep: Callable[["Tweaker", Problem, List[ImmutableArg]], bool] = lambda _t, _p, _a: True
    reword: Optional[Callable[["Tweaker", Problem], str]] = None

    def match(self, problem: Problem) -> Match[str]:
        match = self.pattern.match(problem.text)
        assert match is not None
        return match

    def should_keep(self, problem: Problem, args: List[ImmutableArg]) -> bool:
        return self.keep(self, problem, args)

    def does_reword(self) -> bool:
        return self.reword is not None

    def get_reword(self, problem: Problem) -> str:
        return self.reword(self, problem) if self.reword else problem.text


def invalid_name_keep(self: Tweaker, problem: Problem, args: List[ImmutableArg]) -> bool:
    match = self.match(problem)
    if match.group(1).lower() == "module":
        return False

    name = match.group(2)
    style = match.group(3)
    return style != "snake_case naming style" \
        or any(ch1.islower() and ch2.isupper() for ch1, ch2 in zip(name, name[1:]))


def disallowed_name_keep(self: Tweaker, problem: Problem, args: List[ImmutableArg]) -> bool:
    if len(args) == 1 and args[0].option == Option.ALLOWED_ONECHAR_NAMES and args[0].val is not None:
        assert isinstance(args[0].val, str)
        allowed_onechar_names = args[0].val
    else:
        allowed_onechar_names = ""

    match = self.match(problem)
    name = match.group(1)

    return len(name) != 1 or name not in allowed_onechar_names


def disallowed_name_reword(self: Tweaker, problem: Problem) -> str:
    name = self.match(problem).group(1)
    if len(name) == 1:
        return f"Disallowed single-character variable name \"{name}\", choose a more descriptive name"
    return problem.text


def consider_using_in_reword(self: Tweaker, problem: Problem) -> str:
    match = self.match(problem)

    groups = match.groups()
    start = groups[0]
    outer_quote = groups[1]
    vals = groups[4].split(", ")
    assert vals

    if all(v and v[0] == v[-1] and v[0] in "\"\'" and len(v) == 3 for v in vals):
        inner_quote = vals[0][0]
        return start + inner_quote + "".join(v.strip("\"\'") for v in vals) + inner_quote + outer_quote

    return problem.text


def _get_if_condition(filename: str, line: int) -> str:
    line -= 1  # -1 adjusts for indexing from 0
    with open(filename) as f:
        lines = f.readlines()
        result = [lines[line].strip()]
        while result[-1][-1] == "\\":
            result[-1] = result[-1].strip("\\ ")
            line += 1
            result.append(lines[line].strip())
        return " ".join(result)


def simplifiable_if_statement_reword(self: Tweaker, problem: Problem) -> str:
    if_line = _get_if_condition(problem.path, problem.line)
    if not if_line.startswith("if "):
        return problem.text
    if_line = if_line[len("if "):].strip(":")

    match = self.match(problem)
    return problem.text[:match.start(1)] + if_line + problem.text[match.end(1):]


def _get_if_expression(problem: Problem) -> str:
    with open(problem.path) as f:
        selected_lines = f.readlines()[problem.line - 1:problem.end_line]
    selected_lines[0] = selected_lines[0][problem.column:]  # no -1, columns indexed from 0

    if problem.line == problem.end_line:
        assert problem.end_column is not None
        selected_lines[-1] = selected_lines[-1][:problem.end_column - problem.column]
    else:
        selected_lines[-1] = selected_lines[-1][:problem.end_column]

    return " ".join(line.strip(" \t\n\\") for line in selected_lines)


def simplifiable_if_expression_reword(self: Tweaker, problem: Problem) -> str:
    if_expr = _get_if_expression(problem)
    if_expr_match = re.match(r"(?:True|False) *if *(.*?) *else *(?:True|False)", if_expr)
    if not if_expr_match:
        return problem.text

    match = self.match(problem)
    return problem.text[:match.start(1)] + if_expr_match.group(1) + problem.text[match.end(1):]


def unused_import_keep(self: Tweaker, problem: Problem, args: List[ImmutableArg]) -> bool:
    match = self.match(problem)
    return not match.group(1).startswith("ib111")


Tweakers = Dict[Tuple[Linter, str], Tweaker]

TWEAKERS = {
    (Linter.PYLINT, "C0103"): Tweaker(  # invalid-name
        set(),
        re.compile(r"^(.*) name \"(.*)\" doesn't conform to (.*)$"),
        invalid_name_keep
    ),
    (Linter.PYLINT, "C0104"): Tweaker(  # disallowed-name
        set([Option.ALLOWED_ONECHAR_NAMES]),
        re.compile(r"Disallowed name \"(.*)\""),
        disallowed_name_keep,
        disallowed_name_reword
    ),
    (Linter.PYLINT, "R1714"): Tweaker(  # consider-using-in
        set(),
        re.compile(
            r"^(Consider merging these comparisons with \"in\" to "
            r"(\"|\')([^\s]*)( not)? in )\(([^\)]+)\)(\"|\')"
        ),
        reword=consider_using_in_reword
    ),
    (Linter.PYLINT, "R1703"): Tweaker(  # simplifiable-if-statement
        set(),
        re.compile(
            r"^The if statement can be replaced with '.*?((?:bool\()?test\)?)'"
        ),
        reword=simplifiable_if_statement_reword
    ),
    (Linter.PYLINT, "R1719"): Tweaker(  # simplifiable-if-expression
        set(),
        re.compile(
            r"^The if expression can be replaced with '.*?((?:bool\()?test\)?)'"
        ),
        reword=simplifiable_if_expression_reword
    ),
    (Linter.FLAKE8, "F401"): Tweaker(  # module imported but unused
        set(),
        re.compile("'(.*)' imported but unused"),
        unused_import_keep
    )
}


def get_tweakers() -> Tweakers:
    return TWEAKERS



from edulint.config.config import Config, get_config, get_cmd_args
from edulint.linting.problem import Problem
from edulint.linting.linting import lint_many, sort
from typing import List, Tuple
import argparse
import os


def setup_argparse() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Lints provided code.")
    parser.add_argument("-o", "--option", dest="options", default=[], action="append", help="possible option")
    parser.add_argument("files", metavar="FILE", nargs="+", help="the file(s) to lint")
    parser.add_argument("--json", action="store_true",
                        help="should output problems in json format")
    return parser.parse_args()


def partition(files: List[str], configs: List[Config]) -> List[Tuple[List[str], Config]]:
    dedup_configs = list(set(configs))
    indices = [dedup_configs.index(config) for config in configs]
    partitioned: List[List[str]] = [[] for _ in dedup_configs]

    for i, filename in enumerate(files):
        partitioned[indices[i]].append(filename)

    return list(zip(partitioned, dedup_configs))


def main() -> int:
    args = setup_argparse()
    cmd_args = get_cmd_args(args)

    configs = [get_config(filename, cmd_args=cmd_args) for filename in args.files]

    result = sort(args.files, lint_many(partition(args.files, configs)))
    if args.json:
        print(Problem.schema().dumps(result, indent=2, many=True))  # type: ignore
    else:
        prev_problem = None
        for problem in result:
            if len(args.files) > 1 and (prev_problem is None or prev_problem.path != problem.path):
                print(f"****************** {os.path.basename(problem.path)}")
                prev_problem = problem

            print(problem)
    return 0



from setuptools import setup
import os

# version parsing from __version__ taken from
# https://packaging.python.org/en/latest/guides/single-sourcing-package-version/


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


if __name__ == "__main__":
    setup(version=get_version(os.path.join("edulint", "__init__.py")))
